# Copyright 2014-2015 Jan Decaluwe
#
# This file is part of Urubu.
#
# Urubu is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Urubu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Urubu.  If not, see <http://www.gnu.org/licenses/>.

import posixpath
import re

import markdown
import logging
logging.captureWarnings(False)

from markdown import Extension
from markdown.extensions import toc 
from markdown.treeprocessors import Treeprocessor
from markdown.inlinepatterns import ReferenceInlineProcessor, REFERENCE_RE
from markdown.inlinepatterns import AsteriskProcessor, EmStrongItem 
from markdown.inlinepatterns import EM_STRONG2_RE, STRONG_EM2_RE
from markdown.inlinepatterns import SMART_STRONG_EM_RE, SMART_STRONG_RE, SMART_EMPHASIS_RE

from urubu import UrubuWarning, urubu_warn, UrubuError, _warning, _error


def _set_dl_class(tree):
    for item in tree:
        if item.tag == 'dl':
            item.set('class', 'dl-horizontal')
        _set_dl_class(item)


class DLClass(Treeprocessor):

    def run(self, root):
        _set_dl_class(root)
        return None


class DLClassExtension(Extension):

    """Add 'dl-horizontal' class to definition list elements (for bootstrap)."""

    def extendMarkdown(self, md):
        md.treeprocessors.register(DLClass(md), 'dlclass', 3)


def _set_table_class(tree):
    for item in tree:
        if item.tag == 'table':
            item.set('class', 'table')
        _set_table_class(item)


class TableClass(Treeprocessor):

    def run(self, root):
        _set_table_class(root)
        return None


class TableClassExtension(Extension):

    """Add 'table' class to table elements (for bootstrap)."""

    def extendMarkdown(self, md):
        md.treeprocessors.register(TableClass(md), 'tableclass', 4)


class ProjectReferenceInlineProcessor(ReferenceInlineProcessor):

    def handleMatch(self, m, data):

        text, index, handled = self.getText(data, m.end(0))
        if not handled:
            return None, None, None

        ref, end, shortref, handled = self.evalRef(data, index, text)
        if not handled:
            return None, None, None

        # Clean up linebreaks in ref
        ref = self.NEWLINE_CLEANUP_RE.sub(' ', ref)

        id = ref.lower()

        if id in self.md.references:
            href, title = self.md.references[id]
        else:
            anchor = None
            if '#' in ref:
                ref, anchor = ref.split('#', 1)
            this = self.md.this
            if not posixpath.isabs(ref):
                # treat empty ref as reference to current page
                if not ref: 
                    ref = this['components'][-1]
                rootrelpath = '/' + '/'.join(this['components'][:-1])
                id = posixpath.normpath(posixpath.join(rootrelpath, ref))
                id = id.lower()
            else:
                id = ref.lower()
            ref = ref.lower()
            if ref in self.md.site['reflinks']:
                if (ref != id) and (id in self.md.site['reflinks']):
                    raise UrubuError(_error.ambig_ref_md, msg=ref, fn=this['fn'])
                id = ref
            if id in self.md.site['reflinks']:
                item = self.md.site['reflinks'][id]
                href, title = item['url'], item['title']
                if shortref:
                    text = title
                    if anchor is not None:
                        text = anchor
                if anchor is not None:
                    anchor = toc.slugify(anchor, '-')
                    href = '%s#%s' % (href, anchor)
                    anchorref = '%s#%s' % (id, anchor)
                    self.md.this['_anchorrefs'].add(anchorref)

            else:  # ignore undefined refs
                urubu_warn(_warning.undef_ref_md, msg=ref, fn=this['fn'])
                return None, None, None

        return self.makeTag(href, title, text), m.start(0), end

    def evalRef(self, data, index, text):
        """
        Evaluate ref from [text][ref] or [ref][] 

        """
        shortref = False 
        m = self.RE_LINK.match(data, pos=index)
        if not m:
            return None, index, shortref, False 
        else:
            ref = m.group(1)
            end = m.end(0)
            if not ref:
                ref = text
        return ref, end, shortref, True 

class ProjectShortReferenceInlineProcessor(ProjectReferenceInlineProcessor):

    def evalRef(self, data, index, text):
        """
        Evaluate ref from [text]

        """
        shortref = True 
        return text, index, shortref, True 


class ProjectReferenceExtension(Extension):

    """Overwrite reference patterns with project reference extensions."""

    def extendMarkdown(self, md):
        md.inlinePatterns.deregister('reference')
        md.inlinePatterns.register(ProjectReferenceInlineProcessor(REFERENCE_RE, md), 'reference', 170)
        md.inlinePatterns.deregister('short_reference')
        md.inlinePatterns.register(ProjectShortReferenceInlineProcessor(REFERENCE_RE, md), 'short_reference', 130)


class ExtractAnchorsClass(Treeprocessor):

    def run(self, tree):
        this = self.md.this
        thisid = this['id']
        components = this['components']
        for item in tree:
            if 'id' in item.attrib:
                self.md.anchors.add("%s#%s" % (thisid, item.attrib['id']))
                # add special version for index files
                if components[-1] == 'index':
                    navid = thisid[:-6] # remove trailing backslash also
                    self.md.anchors.add("%s#%s" % (navid, item.attrib['id']))
        return None


class ExtractAnchorsExtension(Extension):

    def extendMarkdown(self, md):
        md.treeprocessors.register(ExtractAnchorsClass(md), 'extractanchors', 5)


# extension for the <mark> tag
class UnderscoreMarkProcessor(AsteriskProcessor):
    """Emphasis processor for handling strong and mark matches inside underscores."""

    PATTERNS = [
        EmStrongItem(re.compile(EM_STRONG2_RE, re.DOTALL | re.UNICODE), 'double', 'strong,mark'),
        EmStrongItem(re.compile(STRONG_EM2_RE, re.DOTALL | re.UNICODE), 'double', 'mark,strong'),
        EmStrongItem(re.compile(SMART_STRONG_EM_RE, re.DOTALL | re.UNICODE), 'double2', 'strong,mark'),
        EmStrongItem(re.compile(SMART_STRONG_RE, re.DOTALL | re.UNICODE), 'single', 'strong'),
        EmStrongItem(re.compile(SMART_EMPHASIS_RE, re.DOTALL | re.UNICODE), 'single', 'mark')
    ]

class MarkTagExtension(Extension):

    def extendMarkdown(self, md):
        md.inlinePatterns.deregister('em_strong2')
        md.inlinePatterns.register(UnderscoreMarkProcessor(r'_'), 'em_strong2', 50)






# Copyright 2015 Sreepathi Pai
#
# This file is part of Urubu.
#
# Urubu is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Urubu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Urubu.  If not, see <http://www.gnu.org/licenses/>.

import http.server as httpserver

class AliasingHTTPRequestHandler(httpserver.SimpleHTTPRequestHandler):
    def do_GET(self):
        baseurl = self.server.baseurl

        if not baseurl:
            httpserver.SimpleHTTPRequestHandler.do_GET(self)
            return

        wk_baseurl = "/%s/" % (baseurl)

        if self.path == wk_baseurl[:-1]:
            # handle /$baseurl -> /$baseurl/

            self.send_response(301, 'Moved Permanently')
            self.send_header('Location', wk_baseurl)
            self.end_headers()
            return

        if self.path[:len(wk_baseurl)] == wk_baseurl:
            # translate /$baseurl/path internally to /path and serve

            self.path = self.path[len(wk_baseurl)-1:]
            httpserver.SimpleHTTPRequestHandler.do_GET(self)
            return
        else:
            # handle /xyz/path -> /$baseurl/xyz/path
            #
            # usually caused by underlying server sending a redirect
            # to non-baseurl-prefixed path

            self.send_response(302, 'Moved Temporarily')
            sep = "/" if self.path[0] != "/" else ""

            # note this replicates underlying bugs in that Location is
            # not absolute as required by the spec and we throw away
            # '?' and '#'

            self.send_header('Location', "/%s%s%s" % (baseurl, sep, self.path))
            self.end_headers()
            return



# Copyright 2014 Jan Decaluwe
#
# This file is part of Urubu.
#
# Urubu is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Urubu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Urubu.  If not, see <http://www.gnu.org/licenses/>.

import os

siteinfofn = '_site.yml'
sitedir = '_build'
tagdir = 'tag'
tagid = '/' + tagdir
tagindexid = tagid + '/' + 'index'
tag_layout = 'tag'
layoutdir = '_layouts'
tipuesearchdir = 'tipuesearch'
tipuesearch_content = 'tipuesearch_content.json'



# Copyright 2014, 2015 Jan Decaluwe
#
# This file is part of Urubu.
#
# Urubu is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Urubu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Urubu.  If not, see <http://www.gnu.org/licenses/>.

import os, sys
import yaml
import fnmatch
import shutil
import datetime
import itertools
import importlib
from operator import itemgetter

from urubu import UrubuWarning, UrubuError, urubu_warn, _warning, _error
from urubu import readers, processors

from urubu.config import (siteinfofn, sitedir,
                          tagdir, tagid, tagindexid, tag_layout)

def require_key(key, mapping, tipe, fn):
    type_error = "{}: '{}' value should be of type {}"
    if key in mapping:
        if not (isinstance(mapping[key], tipe)):
            raise TypeError(type_error.format(fn, key, tipe))


def get_relpath(path, start):
    return os.path.relpath(path, start)


def get_components(path, hasext=True):
    p = path
    if hasext:
        p, ext = os.path.splitext(path)
    p = p.lstrip(os.curdir)
    p = p.strip(os.sep)
    components = []
    if p:
        components = p.split(os.sep)
    return components

def make_id(components, lowercased=True):
    id = ('/' + '/'.join(components))
    if lowercased:
        return id.lower()
    return id


def make_clean(dir):
    for fn in os.listdir(dir):
        p = os.path.join(dir, fn)
        if os.path.isdir(p) and not fn == '.git':
            shutil.rmtree(p)
        elif os.path.isfile(p) and not fn == 'CNAME':
            os.remove(p)


class Project(object):

    def __init__(self):
        self.cwd = os.getcwd()
        self.site = {'baseurl': None,
                     'reflinks': {},
                     'link_ext': '.html',
                     'file_ext': '.html',
                     'mark_tag_support': True
                     }
        self.sitedir = sitedir
        self.get_siteinfo()

        """Get user-defined python hooks."""
        # load _python module from cwd 
        try:
            _python = importlib.import_module('_python')
        except ModuleNotFoundError:
            _python = None
        self.filters = getattr(_python, 'filters', {})
        self.validators = getattr(_python, 'validators', {})
        # overwrite placeholder method if function found
        self.process_info = getattr(_python, 'process_info', self.process_info)

        self.filelist = []
        self.navlist = []
        self.taglist = []
        self.tagmap = {}
        self.layouts = []
        # anchors to be filled in by markdown processor
        self.anchors = set()

    def process_info(self, info, site):
        """Plugin placeholder"""
        pass

    def get_siteinfo(self):
        """Get the siteinfo from the yaml data file."""
        if not os.path.isfile(siteinfofn):
            return
        with open(siteinfofn, encoding='utf-8-sig') as f:
            meta = yaml.safe_load(f)
        # validate the site reflinks and add them
        if 'reflinks' in meta:
            for id in meta['reflinks']:
                info = meta['reflinks'][id]
                self.validate_sitereflink(id, info)
                self.add_reflink(id, info)
            del meta['reflinks']
        self.site.update(meta)
        # check for custom build directory 
        if 'build_dir' in meta:
            builddir = os.path.expanduser(meta['build_dir'])
            self.sitedir = os.path.abspath(builddir)

    def validate_sitereflink(self, id, info):
        if 'title' not in info:
            raise UrubuError(_error.undef_reflink_key, fn=id, msg=id + ': title')
        if 'url' not in info:
            raise UrubuError(_error.undef_reflink_key, fn='_site.yml', msg=id + ': url')

    def add_reflink(self, id, info):
        """Add a valid reflink to the site reflinks."""
        id = id.lower()
        if id in self.site['reflinks']:
            cFile = self.site['reflinks'][id]
            errmsg = "id '{ID}' in {FN} conflicts with {CFN}".\
                format(ID=id, FN=info['fn'], CFN=cFile['fn'])
            raise UrubuError(_error.ambig_refid, msg=errmsg)
        self.site['reflinks'][id] = info

    def finalize_local_url(self, url):
        """Add a base to a local URL, if configured."""
        if self.site['baseurl']:
            url = '/' + self.site['baseurl'] + url
        return url

    def get_contentinfo(self):
        """Get info from the markdown content files."""
        pattern = '*.md'
        ignore_patterns = self.get_ignore_patterns()
        for path, dirnames, filenames in os.walk(self.cwd):
            relpath = get_relpath(path, self.cwd)
            if any(fnmatch.fnmatch(relpath, ip) for ip in ignore_patterns):
                continue

            content_found = index_found = False
            for fn in filenames:
                if fnmatch.fnmatch(fn, pattern):
                    # normalize to convert ./foo into foo
                    # to avoid problems with ignore_patterns matching
                    relfn = os.path.normpath(os.path.join(relpath, fn))
                    if any(fnmatch.fnmatch(relfn, ip) for ip in ignore_patterns):
                        continue
                    meta = readers.get_yamlfm(relfn)
                    if meta is None:
                        urubu_warn(_warning.no_yamlfm, fn=relfn)
                        continue
                    fileinfo = self.make_fileinfo(relfn, meta)
                    self.filelist.append(fileinfo)
                    self.process_info(fileinfo, self.site)
                    # validate after file info has been added so it can be used
                    self.validate_fileinfo(fileinfo)
                    self.add_reflink(fileinfo['id'], fileinfo)
                    if fn == 'index.md':
                        index_found = True
                        # start from fileinfo of index file
                        navinfo = self.make_navinfo(relpath, fileinfo)
                        self.navlist.append(navinfo)
                        self.validate_navinfo(navinfo)
                        self.add_reflink(navinfo['id'], navinfo)
                        # add nav info to tag map
                        self.add_info_to_tagmap(navinfo)
                    else:
                        content_found = True
                        # add id for non-index files to tag tags
                        self.add_info_to_tagmap(fileinfo)
            # a folder with content but no index is an error
            if content_found and not index_found:
                raise UrubuError(_error.no_index, msg='', fn=relpath)

    def validate_fileinfo(self, info):
        fn = info['fn']
        # layout is mandatory
        if 'layout' not in info:
            raise UrubuError(_error.undef_info, msg='layout', fn=fn)
        if info['layout'] is None:
            return
        layout = info['layout']
        # modification date, always available
        t = os.path.getmtime(fn)
        info['mdate'] = datetime.date.fromtimestamp(t)
        # first run a validator if it exist
        if layout in self.validators:
            self.validators[layout](info)
        # a validator may add/modify attributes
        layout = info['layout']
        if layout is None:
            return
        if layout not in self.layouts:
            self.layouts.append(layout)
        # title
        if 'title' not in info:
            raise UrubuError(_error.undef_info, msg='title', fn=fn)
        # Support integer titles by converting to string
        # this matters for the json dump for search
        # test type explicitly to avoid problems with unicode titles...
        if isinstance(info['title'], int):
            info['title'] = str(info['title'])
        # date
        if 'date' in info:
            if not isinstance(info['date'], datetime.date):
               raise UrubuError(_error.date_format, fn=fn)
        # tags
        if 'tags' in info:
            # TODO: make sure it's a list of strings
            if isinstance(info['tags'], str):
                info['tags'] = [info['tags']]
            # TODO: normalize tags

    def validate_navinfo(self, info):
        fn = info['indexfn']
        if ('content' not in info) and ('order' not in info):
            # exception: tag folder doesn't require content atribute
            # set it to empty list align with normal folders
            if info['id'] == tagid:
                info['content'] = []
            else:
                raise UrubuError(_error.undef_content, fn=fn)
        require_key('content', info, list, fn)

    def make_fileinfo(self, relfn, meta):
        """Make a fileinfo dict."""
        info = {}
        info['fn'] = relfn
        info['components'] = components = get_components(
            relfn if meta.get('saveas') is None else meta.get('saveas'))
        info['id'] = make_id(components)
        # make html url from ref
        info['url'] = self.finalize_local_url(
            make_id(components, lowercased=False) + self.site['link_ext'])
        info['_anchorrefs'] = set()
        info.update(meta)
        return info

    def make_navinfo(self, relpath, fileinfo):
        """Make a navinfo dict."""
        # start from fileinfo of index file
        info = fileinfo.copy()
        # overwrite attributes according to navinfo view
        info['indexfn'] = info['fn']
        info['fn'] = relpath
        info['components'] = components = get_components(relpath)
        info['id'] = make_id(components)
        # add trailing slash for navigation url
        info['url'] = self.finalize_local_url(info['id'])
        if info['url'] != '/':
            info['url'] += '/'
        return info

    def add_info_to_tagmap(self, info):
        """Add id to tagmap."""
        # tags are optional
        if 'tags' not in info:
            return
        for tag in info['tags']:
            if tag not in self.tagmap:
                self.tagmap[tag] = []
            self.tagmap[tag].append(info)

    def resolve_reflinks(self):
        for info in self.navlist:
            if 'content' in info:
                self.resolve_content(info)
            else:  # order
                assert 'order' in info
                self.get_content(info)
            # make reflinks content available in index file also
            index_id = make_id(info['components'] + ['index'])
            self.site['reflinks'][index_id]['content'] = info['content']

    def resolve_content(self, info):
        """Resolve reflinks in a folder."""
        refcontent = []
        for item in info['content']:
            if isinstance(item, dict):
                link = self.resolve_linkspec(item, info)
            else:
                link = self.resolve_ref(item, info)
            refcontent.append(link)
        # overwrite content ids with resolved refs
        info['content'] = refcontent

    def resolve_ref(self, ref, info):
        """Resolve a reference."""
        reflinks = self.site['reflinks']
        ref = ref.lower()
        path = os.path.normpath(os.path.join(info['fn'], ref))
        indexfn = info['indexfn']
        id = make_id(get_components(path, hasext=False))
        if ref in reflinks:
            if (ref != id) and (id in reflinks):
                raise UrubuError(_error.ambig_ref, msg=ref, fn=indexfn)
            id = ref
        elif not id in reflinks:
            raise UrubuError(_error.undef_ref, msg=ref, fn=indexfn)
        return reflinks[id]

    def resolve_linkspec(self, linkspec, info):
        link = {}
        link['url'] = link['title'] = linkspec.get('url', None)
        if 'ref' in linkspec:
            reflink = self.resolve_ref(linkspec['ref'], info)
            link.update(reflink)
        if 'title' in linkspec:
            link['title'] = linkspec['title']
        return link

    def get_content(self, info):
        """Infer sorted content of a folder."""
        reflinks = self.site['reflinks']
        refcontent = []
        key = info['order']
        reverse = info.get('reverse', False)
        navcomps = info['components']

        def pred(item):
            itemcomps = item['components']
            if len(itemcomps) == len(navcomps) + 1 and \
               itemcomps[:-1] == navcomps and \
               itemcomps[-1] != 'index' and \
               item['layout'] is not None:
                if key not in item:
                    raise UrubuError(_error.undef_key, msg=key, fn=item['fn'])
                return True
            return False
        allinfo = itertools.chain(self.filelist, self.navlist)
        refcontent = filter(pred, allinfo)

        def get_keyval(item):
            return item[key]
        refcontent = sorted(refcontent, key=get_keyval, reverse=reverse)
        info['content'] = refcontent

    def make_breadcrumbs(self):
        for info in self.filelist:
            breadcrumbs = []
            id = ''
            comps = info['components']
            # discard index
            if comps[-1] == 'index':
                comps = comps[:-1]
            for comp in comps:
                id = id + '/' + comp
                id = id.lower()
                breadcrumbs.append(self.site['reflinks'][id])
            info['breadcrumbs'] = breadcrumbs

    def make_pager(self):
        for info in self.navlist:
            content = info['content']
            if not content:
                continue
            content[0]['prev'] = None
            for i in range(1, len(content)):
                content[i - 1]['next'] = content[i]
                content[i]['prev'] = content[i - 1]
            content[-1]['next'] = None

    def make_taginfo(self, tag, content):
        """Make a taginfo dict."""
        info = {}
        info['title'] = info['tag'] = tag
        info['layout'] = tag_layout
        info['fn'] = os.path.join(tagdir, tag, 'index')
        info['components'] = components = (tagdir, tag)
        info['id'] = make_id(components, lowercased=False)
        # add trailing slash for tag index url
        info['url'] = self.finalize_local_url(info['id']) + '/'
        info['content'] = content
        return info

    def process_tags(self):
        """ Process tag map and tag content."""
        def get_date(item):
            """ Return item's date, or mdate as fallback."""
            if 'date' in item:
                return item['date']
            else:
                return item['mdate']
        for tag in self.tagmap.keys():
            # sort tag content by date
            content = sorted(self.tagmap[tag], key=get_date, reverse=True)
            taginfo = self.make_taginfo(tag, content)
            self.taglist.append(taginfo)
            self.add_reflink(taginfo['id'], taginfo)

        # sort tags according to content length, then alphabetically
        # to get sort order right, invert content length
        def get_tagkey(taginfo):
            return (-len(taginfo['content']), taginfo['tag'])
        self.taglist = sorted(self.taglist, key=get_tagkey)

        # set up tagid info dict if it doesn't exist already
        if tagid not in self.site['reflinks']:
            self.site['reflinks'][tagid] = {}
        self.site['reflinks'][tagid]['content'] = self.taglist
        # propagate content to index file if it exists
        if tagindexid in self.site['reflinks']:
            self.site['reflinks'][tagindexid]['content'] = self.taglist

    def get_ignore_patterns(self):
        ignore_patterns = ('.?*', '_*', 'Makefile')
        if 'ignore_patterns' in self.site:
            ignore_patterns += tuple(self.site['ignore_patterns'])
        return ignore_patterns

    def get_keep_files(self):
        if 'keep_files' in self.site:
            return self.site['keep_files']
        return []

    def check_anchor_links(self):
        for info in self.filelist:
            for ar in info['_anchorrefs']:
                if not ar in self.anchors:
                    urubu_warn(_warning.undef_anchor, msg=ar, fn=info['id'] )

    def make_site(self):
        """Make the site."""
        # Keep sitedir alive if it exists, for the server
        if not os.path.exists(self.sitedir):
            os.mkdir(self.sitedir)
        make_clean(self.sitedir)
        ignore_patterns = self.get_ignore_patterns() + ('*.md',)
        ignore = shutil.ignore_patterns(*ignore_patterns)
        for fn in os.listdir(self.cwd):
            if any(fnmatch.fnmatch(fn, ip) for ip in ignore_patterns):
                continue
            wp = os.path.join(self.cwd, fn)
            sp = os.path.join(self.sitedir, fn)
            if os.path.isdir(wp):
                shutil.copytree(wp, sp, ignore=ignore)
            elif os.path.isfile(wp):
                shutil.copyfile(wp, sp)

        # explicit files to keep
        for fn in self.get_keep_files():
            wp = os.path.join(self.cwd, fn)
            sp = os.path.join(self.sitedir, fn)
            if os.path.isfile(wp):
                shutil.copyfile(wp, sp)

        # make tag index dirs
        if self.taglist:
            tagpath = os.path.join(self.sitedir, tagdir)
            if not os.path.isdir(tagpath):
                os.mkdir(tagpath)
            for taginfo in self.taglist:
                os.mkdir(os.path.join(tagpath, taginfo['tag']))
        self.process_content()
        self.check_anchor_links()

    def process_content(self):
        """Process the content files."""
        p = processors.ContentProcessor(self.sitedir, project=self)
        p.process()

def load():
    proj = Project()
    return proj

def build():
    proj = load()
    proj.get_contentinfo()
    proj.resolve_reflinks()
    proj.make_breadcrumbs()
    proj.make_pager()
    proj.process_tags()
    proj.make_site()



# Copyright 2014 Jan Decaluwe
#
# This file is part of Urubu.
#
# Urubu is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Urubu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Urubu.  If not, see <http://www.gnu.org/licenses/>.

import os, sys, json, itertools

import markdown
import logging
logging.captureWarnings(False)
from bs4 import BeautifulSoup

import jinja2

from urubu import UrubuWarning, UrubuError, urubu_warn, _warning
from urubu import md_extensions

from urubu.config import layoutdir, tag_layout, tipuesearchdir, tipuesearch_content

def skip_yamlfm(f):
    """Return source of a file without yaml frontmatter."""
    f.readline()
    found = False
    lines = []
    for line in f.readlines():
        if found:
            lines.append(line)
        if line.strip() == '---':
            found = True
    return ''.join(lines)


class ContentProcessor(object):

    def __init__(self, sitedir, project):
        self.sitedir = sitedir
        self.filelist = project.filelist
        self.navlist = project.navlist
        self.taglist = project.taglist
        self.site = project.site
        dlclass = md_extensions.DLClassExtension()
        tableclass = md_extensions.TableClassExtension()
        projectref = md_extensions.ProjectReferenceExtension()
        extractanchors = md_extensions.ExtractAnchorsExtension()
        marktag = md_extensions.MarkTagExtension()
        # there is a strange interaction between smarty and reference links that start on a new line
        # disabling smarty for now...
        # extensions = ['extra', 'codehilite', 'headerid', 'toc', 'smarty', tableclass, projectref]
        extensions = ['markdown.extensions.extra', 'markdown.extensions.codehilite', 'markdown.extensions.toc',
                      dlclass, tableclass, projectref, extractanchors]
        if self.site['mark_tag_support']:
            extensions.append(marktag)
        extension_configs = {'markdown.extensions.codehilite': [('guess_lang', 'False'),
                                                                ('linenums', 'False')],
                             'markdown.extensions.toc': [('baselevel', 2)]
                             }
        self.md = markdown.Markdown(extensions=extensions,
                                    extension_configs=extension_configs)
        self.md.site = self.site
        self.md.anchors = project.anchors
        if 'strict_undefined' in self.site and self.site['strict_undefined']:
            undefined_class = jinja2.StrictUndefined
        else:
            undefined_class = jinja2.Undefined
        env = self.env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(layoutdir),
            lstrip_blocks=True,
            trim_blocks=True,
            undefined=undefined_class)
        env.filters.update(project.filters)
        self.templates = {}
        for layout in project.layouts:
            self.templates[layout] = self.env.get_template(layout + '.html')
        # layout for tags is optional, triggers index file generation per tag
        try:
            self.templates[tag_layout] = self.env.get_template(
                tag_layout + '.html')
        except jinja2.exceptions.TemplateNotFound:
            if self.taglist:
                urubu_warn(_warning.undef_tag_layout, msg=tag_layout)

    def process(self):
        """Process the content.

        Conversion and rendering are done in separate phases, so
        that the full content is available to the rendering process.
        """
        self.convert()
        self.alt_layouts()
        self.pagination()
        self.render()
        self.make_tipuesearch_content()

    def alt_layouts(self):
        # If a file contains the alt_layouts attribute, then
        # we need to generate a copy of the file for each
        # additional layout, in the location given.  Each 
        # alt layout has two fields, layout and location
        for info in self.filelist:
            if 'alt_layouts' in info: 
                for layout in info['alt_layouts']:
                    new_info = info.copy()
                    
                    # Make sure the alt layout is available
                    if layout['layout'] not in self.templates:
                        self.templates[layout['layout']] = self.env.get_template(layout['layout'] + '.html')
                        
                    # Set the layout of the new file
                    new_info['layout'] = layout['layout']
                    
                    # Set path of the new file
                    import ntpath
                    filename = ntpath.basename(new_info['fn'])
                    new_info['fn'] = layout['location']+"\\"+filename                   
                    
                    # Prevent the new file from generating more new files
                    del new_info['alt_layouts']
                    
                    self.filelist.append(new_info)
        
    def pagination(self):
        # We look at attributes in pages here, e.g.:
        #
        #   items_per_page: 5
        #   items_index: news\index.md
        #
        #   OR
        #
        #   items_per_page: 5
        #   items_filter: mysteries time-loop
        #
        # We then find the info item for the items index/filter and
        # use it to create more entries for the current item in
        # the file list, splitting it out into items_per_page
        # chunks for each file.
        #  
        # The filters are the same ones defined in _python for Jinja2
        # to use.  The name of the function is followed by its parameters
        #
        # The page and any new pages created from it will get new attributes:
        #
        #     numpages - the number of total pages generated
        #     thispage - the page number of this page in the chain
        #     prevpage - the previous page in the chain
        #     nextpage - the next page in the chain
        #     pages    - a list of page numbers and page objects
        # 
        # The first page will have no prevpage, and the last page
        # will have no nextpage.

        import math
        
        for info in self.filelist:
            source = None
            
            if 'items_per_page' in info:
                if 'items_index' in info:
                    if info['items_index'] == "this":
                        source = info['content']
                    else:
                        source = next(x for x in self.filelist if x['fn'] == info['items_index'])['content']
            
                if 'items_filter' in info:
                    filter = info['items_filter'].split()
                    source = self.env.filters[filter[0]](*filter[1:])
                
                if source:
                    items_per_page = info['items_per_page']
                    
                    # This will be a shared list among all the pages
                    # that lists each page along with its page number,
                    # e.g. {'pagenum': 1, 'page': info}
                    pages = []
                    
                    # Split source into items_per_page sized chunks,
                    # starting with info and then creating new
                    # files with incrementing numbers
                    
                    chunks = math.ceil(len(source) / items_per_page)

                    # First chunk is always the current page
                    info['content'] = source[0:items_per_page]
                    
                    # See if we even need to worry about pagination
                    # Maybe everything fits on the one page already
                    if len(source) <= items_per_page:
                        continue
                    
                    info['numpages'] = chunks
                    info['thispage'] = 1
                    info['pages'] = pages
                    
                    pages.append({'pagenum': 1, 'page': info})
                    
                    chunk = 1
                    prev_page = info
                    
                    # If we need more chunks, split off new pages for them
                    while chunk < chunks:
                        chunk += 1
                        new_info = info.copy()
                        # Prevent the new page from spinning off new pages
                        del new_info['items_per_page']
                        
                        new_info['pages'] = pages
                        pages.append({'pagenum': chunk, 'page': new_info})
                        
                        # Set the content to the right slice of the source
                        new_info['content'] = source[(chunk-1)*items_per_page:(chunk-1)*items_per_page+items_per_page]
                        
                        new_info['numpages'] = chunks
                        new_info['thispage'] = chunk

                        # Name the new page based on its chunk number
                        fn_parts = os.path.splitext(new_info['fn'])
                        new_info['fn'] = fn_parts[0]+str(chunk)+fn_parts[1]

                        fn_parts = os.path.splitext(new_info['url'])
                        new_info['url'] = fn_parts[0]+str(chunk)+fn_parts[1]
                        
                        # Setup the prevpage and nextpage attributes
                        new_info['prevpage'] = prev_page
                        prev_page['nextpage'] = new_info
                        # We may have inherited this from the first chunk
                        if 'nextpage' in new_info:
                            del new_info['nextpage']
                        
                        prev_page = new_info
                        
                        self.filelist.append(new_info)
        
    def convert(self):
        for info in self.filelist:
            fn = info['fn']
            with open(fn, encoding='utf-8-sig') as inf:
                src = skip_yamlfm(inf)
            self.md.this = info
            # first process as a template
            try:
                templ = self.env.from_string(src)
                src = templ.render(this=info, site=self.site)
            except:
                exc, msg, tb = sys.exc_info()
                raise UrubuError(str(exc), msg=msg, fn=fn)
            self.md.toc = ''
            info['body'] = self.md.convert(src)
            info['toc'] = ''
            if hasattr(self.md, 'toc'):
                # filter out empty tocs, 35 is the magic length
                if len(self.md.toc) > 35:
                    info['toc'] = self.md.toc
            # markdown support in keys
            mdkeys = [key for key in info if key[-3:] == '.md']
            for mdkey in mdkeys:
                key = mdkey[:-3]
                info[key] = self.md.convert(info[mdkey])
            self.md.reset()
        for info in self.navlist:
            # markdown support in keys
            mdkeys = [key for key in info if key[-3:] == '.md']
            for mdkey in mdkeys:
                key = mdkey[:-3]
                info[key] = self.md.convert(info[mdkey])
            self.md.reset()

    def render(self):
        # content files
        for info in self.filelist:
            if info['layout'] is None:
                continue
            self.render_file(info)
        # tag index files
        if tag_layout not in self.templates:
            return
        for info in self.taglist:
            self.render_file(info)

    def render_file(self, info):
        layout = info['layout']
        templ = self.templates[layout]
        html = templ.render(this=info, site=self.site)
        # extract text from html for search support
        self.extract_text(html, info)
        fn = info['fn']

        # check if filename is overriden
        if info.get('saveas') is not None:
            outfn = os.path.join(self.sitedir, info.get('saveas'))
        else:
            bfn, ext = os.path.splitext(fn)
            outfn = os.path.join(self.sitedir, bfn) + self.site['file_ext']

        with open(outfn, 'w', encoding='utf-8', errors='strict') as outf:
            outf.write(html)

    def extract_text(self, html, info):
        # select main tag for search content
        m = BeautifulSoup(html, "html.parser").select('main')
        text = ""
        if m:
            text = m[0].get_text(" ", strip=True)
        info['text'] = text

    def make_tipuesearch_content(self):
        tsd = os.path.join(self.sitedir, tipuesearchdir)
        if not os.path.isdir(tsd):
            return
        tsc = os.path.join(tsd, tipuesearch_content)
        items = []
        # use tag index files if they have been rendered
        taglist = []
        if tag_layout in self.templates:
           taglist = self.taglist
        for info in itertools.chain(self.filelist, taglist):
            if 'text' not in info:
                continue
            tags = ""
            if 'tags' in info:
               tags = ' '.join(info['tags'])
            item = {'text' : info['text'],
                    'title': info['title'],
                    'url'  : info['url'],
                    'tags' : tags}
            items.append(item)
        if len(items) <= 0:
            return
        obj = {'pages': items}
        with open(tsc, 'w', encoding='utf-8') as fd:
            # json.dump is buggy in Python2 -- use workaround
            # print json.dumps(obj, ensure_ascii=False, indent=4)
            data = json.dumps(obj, ensure_ascii=False, indent=4, sort_keys=True)
            fd.write(str(data))



# Copyright 2014 Jan Decaluwe
#
# This file is part of Urubu.
#
# Urubu is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Urubu is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Urubu.  If not, see <http://www.gnu.org/licenses/>.

import yaml
import re

def get_yamlfm(fn):
    """Return the yaml frontmatter."""
    info = _get_yamlfm_helper(fn)
    return info


def _get_yamlfm_helper(fn):
    with open(fn, 'r', encoding='utf-8-sig') as f:
        line = f.readline()
        if line.strip() != '---':
            return None
        lines = []
        while True:
            line = f.readline()
            if not line:
                return None
            elif line.strip() == '---':
                s = ''.join(lines)
                meta = yaml.safe_load(s)
                if isinstance(meta, dict):
                    return meta
                else:
                    return None
            else:
                lines.append(line)





from setuptools import setup

requires = ['jinja2 >= 2.10', 'pygments',
            'markdown >= 3.0','pyyaml', 'beautifulsoup4']

entry_points = {
    'console_scripts': [
        'urubu = urubu.main:main',
    ]
}

setup(
    name="urubu",
    version="1.4.1",
    url='http://urubu.jandecaluwe.com',
    author='Jan Decaluwe',
    author_email='jan@jandecaluwe.com',
    description="A micro CMS for static websites with a focus "
                "on good navigation practices.",
    packages=['urubu'],
    include_package_data=True,
    install_requires=requires,
    entry_points=entry_points,
    extras_require={
        'test': ['pytest', 'sh'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
